"use strict";

/** @typedef {import("./index").WorkerResult} WorkerResult */

/** @typedef {import("./index").FilenameFn} FilenameFn */
const isFilenameProcessed = Symbol("isFilenameProcessed");
/**
 * @template T
 * @param {WorkerResult} result
 * @param {import("./index").InternalWorkerOptions<T>} options
 * @param {undefined | string | FilenameFn} filenameTemplate
 */

function processFilenameTemplate(result, options, filenameTemplate) {
  if ( // @ts-ignore
  !result.info[isFilenameProcessed] && typeof filenameTemplate !== "undefined" && typeof options.generateFilename === "function") {
    result.filename = options.generateFilename(filenameTemplate, {
      filename: result.filename
    });
    result.filename = result.filename.replace(/\[width\]/gi, result.info.width).replace(/\[height\]/gi, result.info.height); // @ts-ignore

    result.info[isFilenameProcessed] = true;
  }
}
/**
 * @template T
 * @param {WorkerResult} result
 * @param {import("./index").InternalWorkerOptions<T>} options
 */


function processSeverityError(result, options) {
  if (options.severityError === "off") {
    result.warnings = [];
    result.errors = [];
  } else if (options.severityError === "warning") {
    result.warnings = [...result.warnings, ...result.errors];
    result.errors = [];
  }
}
/**
 * @template T
 * @param {import("./index").InternalWorkerOptions<T>} options
 * @returns {Promise<WorkerResult>}
 */


async function worker(options) {
  var _result, _result$info, _result2, _result2$errors, _result3, _result3$warnings, _result4, _result4$filename;

  /** @type {WorkerResult} */
  let result = {
    data: options.input,
    filename: options.filename,
    warnings: [],
    errors: [],
    info: {}
  };

  if (!result.data) {
    result.errors.push(new Error("Empty input"));
    return result;
  }

  const transformers = Array.isArray(options.transformer) ? options.transformer : [options.transformer];
  /** @type {undefined | string | FilenameFn} */

  let filenameTemplate;

  for (const transformer of transformers) {
    if (typeof transformer.filter === "function" && !transformer.filter(options.input, options.filename)) {
      continue;
    }
    /** @type {WorkerResult | null} */


    let processedResult;

    try {
      // eslint-disable-next-line no-await-in-loop
      processedResult = await transformer.implementation(result, transformer.options);
    } catch (error) {
      result.errors.push(error instanceof Error ? error : new Error(
      /** @type {string} */
      error));
      return result;
    }

    if (processedResult && !Buffer.isBuffer(processedResult.data)) {
      result.errors.push(new Error("minimizer function doesn't return the 'data' property or result is not a 'Buffer' value"));
      return result;
    }

    if (processedResult) {
      var _filenameTemplate;

      result = processedResult;
      (_filenameTemplate = filenameTemplate) !== null && _filenameTemplate !== void 0 ? _filenameTemplate : filenameTemplate = transformer.filename;
    }
  }

  (_result$info = (_result = result).info) !== null && _result$info !== void 0 ? _result$info : _result.info = {};
  (_result2$errors = (_result2 = result).errors) !== null && _result2$errors !== void 0 ? _result2$errors : _result2.errors = [];
  (_result3$warnings = (_result3 = result).warnings) !== null && _result3$warnings !== void 0 ? _result3$warnings : _result3.warnings = [];
  (_result4$filename = (_result4 = result).filename) !== null && _result4$filename !== void 0 ? _result4$filename : _result4.filename = options.filename;
  processSeverityError(result, options);
  processFilenameTemplate(result, options, filenameTemplate);
  return result;
}

module.exports = worker;